/*
  Toybox beta 1.0 : written by Tse Huong Choo Apr 1993
  beta 1.1 Mon May 24 13:02:43 BST 1993 - Kludge from Hell!
  Version 1.2 Mon May 31 11:20:51 BST 1993 - fixes, enhancements.
  Version (pre release ) 1.3
  Thu Jun  3 17:53:26 BST 1993 - transparent toplevelwindow.

  Version 1.3 (release) Wed Jun 16 13:11:15 BST 1993 - fixed to
  work with ol[v]wm, [v]twm - see README for more info.

  
  Note: The function prototype for XFree() on different machines
  differs. eg void * on mine and char * on some DECs. If you
  have trouble compiling, try changing the overloaded GFree()
  functions so that the locals are XPointer instead of char *.

  If that still doesn't work, look into /usr/include/X11/Xlib.h
  for the prototype of XFree() and use the appropiate <type-name> *.
  
  Email suggestions/kludge-zappers to thc2@doc.ic.ac.uk
  
  Known bugs :
  IF this crashes on startup with an error in XUnmapWindow()
  then examine the definition of unmap_existing_icons() and
  read the comments there.

  Memory leaks - don't have time to find out why yet!
  
  Bevelled edges of the frame surrounding the icon-label doesn't
  appear "popped" out initially when window/icon is created.
  
  Bus errors within XpmReadData() and XtAppAddTimeOut() - bad pointers?
  Xlib fatal error about accessing non-existent windows- specifically:
  audiotool's exit causes an XGetProperty() error within external_events().
  Should be fixed soon.

  And others.

  Requires: xpm.h and libXpm.a to compile (the Xpm-library)
  If you have the SHAPE extension (check /usr/include/X11/extensions
  for the file called shape.h) , then define HAVE_SHAPE_EXTENSION.
  

  TODO:
  Integrate the SHAPE extension bits - it's expected that
  most X servers support the SHAPE extension.

  Make it work with ol[v]wm - almost done.
  
  Must free after call to XGetApplicationResources() ? 
  
  Batch calls to XShape...() to speed up layout - almost done.

  Clean up the exit procedure in RemoveIcon() - windows
  that are mapped should not have their icons re-mapped.

  Find a way to make the closeness-field in the Xpm -attr
  specficiable (the current value  is a subjective guide).
  
  Get rid of strcpy() str...() usage - they may use free and
  malloc which *MAY* intefere with C++'s new/delete.

  Fix potential memory leak in external_events() in
  case PropertyNotify: if( xe->atom == WM_STATE )

  Check layout algorithm - some complaints of it going
  "out of bounds".
  
  additional resources:
  toybox*window-title-15chars-max:  full-pathname-of-xpm-file
  toybox*defaultPixmapFile: full-pathname-of-default-xpm-file

  toybox*window-title-15chars-max: DONT_SHOW
  tells toybox not to produce an icon for that window.
  The semantics when windows change names is still undefined.
  ie I still haven't made up my mind!

  toybox*floatOnRoot: YES
  causes toybox to have a transparent window. Works only
  if toybox is compiled with HAVE_SHAPE_EXTENSION #defined.

  toybox*layoutPolicy: top_left
  or right_bottom or bottom_left etc. but not
  top_top, left_right, bottom_top etc.

  If layoutPolicy is specified as top_left, then
  the icons will remain as much to the "top" as possible before
  having to go "left"
  
  Clicking the right-mouse-button raises the corresp. window,
  mapping it first, if necessary.
  Shift and Button3 on an icon will delete it - you can't get it back!
  
  Note: Toybox is provided "as-is", without any express or implied
  warranty. In no event will the author be held liable for any damage
  arising from the use of this software.

  If you would like include some of your own enhancements/changes,
  please email me at thc2@doc.ic.ac.uk and we could work something
  out.
  

  author's notes:
  direction = 0 - top_left, 1 - top_right, 2 - bottom_left,
  3 - bottom_right, 4 - left_top, 5 - left_bottom,
  6 - right_top, 7 - right_bottom. - enum-ed already.
  */



#include <X11/Xatom.h>
#include <X11/Intrinsic.h>  
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#ifdef HAVE_SHAPE_EXTENSION
#include <X11/extensions/shape.h>
#endif

#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/MessageB.h>

#include <xpm.h>

static char * defaultpixmapdata[] = {
"72 62 3 1",
" 	c #F2F29B9B4444",
".	c #FFFFFFFFFFFF",
"X	c #000033339999",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"      ..............            .......            .....       .....    ",
"      ..    ...   ..          ...     ...            ...       ...      ",
"      .     ...    .         ...       ...           ...       ...      ",
"      .     ...    .         ..         ...           ...     ...       ",
"            ...             ...         ...            ...   ...        ",
"            ...             ...          ..            ...  ...         ",
"            ...            ...           ...            ... ..          ",
"            ...            ...           ...             ....           ",
"            ...            ...           ...             ...            ",
"            ...            ...          ....             ...            ",
"            ...             ...          ..              ...            ",
"            ...             ...         ...              ...            ",
"            ...              ...        ..               ...            ",
"            ...               ..      ...                ...            ",
"          .......              ...   ...               ........         ",
"                                 .....                                  ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"                                                                        ",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXX.........XXXXXXXXXXXXXXXXX.....XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXX..XXXXX..XXXXXXXXXXXXX..XXXXX..XXXXXXXXXXX.....XXXXXXX...XXXXXX",
"XXXXXXXXX..XXXXXX..XXXXXXXXXXX..XXXXXXX..XXXXXXXXXXXX...XXXXXXX..XXXXXXX",
"XXXXXXXXX..XXXXXX..XXXXXXXXXX..XXXXXXXXX..XXXXXXXXXXXX...XXXXX..XXXXXXXX",
"XXXXXXXXX..XXXXXX..XXXXXXXXX...XXXXXXXXX...XXXXXXXXXXXX...XXX..XXXXXXXXX",
"XXXXXXXXX..XXXXX...XXXXXXXXX..XXXXXXXXXXX..XXXXXXXXXXXXX..XXX.XXXXXXXXXX",
"XXXXXXXXX..XXXXX..XXXXXXXXX...XXXXXXXXXXX...XXXXXXXXXXXXX..X..XXXXXXXXXX",
"XXXXXXXXX..XX....XXXXXXXXXX...XXXXXXXXXXX...XXXXXXXXXXXXX....XXXXXXXXXXX",
"XXXXXXXXX..XXXX...XXXXXXXXX...XXXXXXXXXXX...XXXXXXXXXXXXX....XXXXXXXXXXX",
"XXXXXXXXX..XXXXXX..XXXXXXXX...XXXXXXXXXXX...XXXXXXXXXXXX.....XXXXXXXXXXX",
"XXXXXXXXX..XXXXXX...XXXXXXX...XXXXXXXXXXX..XXXXXXXXXXXXX.XX...XXXXXXXXXX",
"XXXXXXXXX..XXXXXXX..XXXXXXXX..XXXXXXXXXXX..XXXXXXXXXXXX..XXX...XXXXXXXXX",
"XXXXXXXXX..XXXXXXX..XXXXXXXX...XXXXXXXXX...XXXXXXXXXXX..XXXXX...XXXXXXXX",
"XXXXXXXXX..XXXXXXX..XXXXXXXXX..XXXXXXXXX..XXXXXXXXXXXX.XXXXXXX...XXXXXXX",
"XXXXXXXXX...XXXXX..XXXXXXXXXXX..XXXXXXX..XXXXXXXXXXX...XXXXXXX...XXXXXXX",
"XXXXXXX...........XXXXXXXXXXXXXX..XXX..XXXXXXXXXXXX.....XXXXX......XXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX...XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"};


#include <iostream.h>
#include <stdlib.h>
#include <string.h>


/*#define DEFAULT_XPM_FILE "/homes/thc2/X11/pixmaps/markoya"*/
#define DEFAULT_XPM_FILE "/opt/toybox/lib/X11/markoya"

#define MAX_NAME_LENGTH 16
#define TIME_OUT 500
#define UNSORTED 0   // shape extension wasn't explicit about the encoding
#define BB_SPACING 6
#define DONT_SHOW "DONT_SHOW"
#define CLOSENESS 22000


extern WidgetClass xmLabelWidgetClass;
extern WidgetClass xmPushButtonWidgetClass;
extern WidgetClass xmFrameWidgetClass;

enum Direction { top_left, top_right, bottom_left, bottom_right,
				 left_top, left_bottom, right_top, right_bottom };

class WM_Icon {
  int done, using_default;
  Widget label, pb, frame, board, board_parent;
  Window w;
  Display *display; // disp the Window w is attached to
  int icon_width, icon_height;
  char *myname, *xpmfilename;
  Boolean to_popout;
public:
  // pixmap to use when all else fails
  static Pixmap defaultpixmap;
  
  WM_Icon() {done = 0; myname = xpmfilename = 0; }
  
  void icon_init( Display *d, Window win, char *name, Widget parent,
				  char *xpmfile ) {
	w = win;
	display = d;
	board_parent = parent;
	// Use String.h??
	//KNF
	myname = new char [ strlen( name )+1 ] ; 
	xpmfilename = new char [ strlen( xpmfile )+1 ] ;
	(void) strcpy( myname, name );
	(void) strcpy( xpmfilename, xpmfile );
	done = 0;

	
  }
  
  void icon_augment_translations( XtTranslations trans ) {
	XtAugmentTranslations( pb, trans );
  }

  void icon_realize();
  
  int is_done() { return done ;  }
  
  void icon_manage(int x, int y );

  Window get_window() { return w; }
  
  void icon_add_activate_callback( XtCallbackProc proc, XtPointer client );

  int is_unmapped();
  
  void icon_bounding_box(int *x_ret, int *y_ret, int *wid, int *hei );
  void icon_pb_bounding_box( int *x_ret, int *y_ret, int *wid, int *hei );
  void icon_frame_bounding_box( int *x_ret, int *y_ret, int *wid, int *hei );
  
  void icon_moveto( int x, int y );
  void icon_set_name( char *n );
  
  void icon_iconify();
  void icon_deiconify();
  
  void icon_popout();
  void icon_popin();
  
  void set_popout_state( Boolean b ) { to_popout = b; }

  ~WM_Icon();

};

void WM_Icon::icon_realize()
{

  XmString s;
  Pixmap pixmap;
  
  board = XtVaCreateWidget( "iconboard", xmBulletinBoardWidgetClass,
							board_parent,
							XmNmarginWidth, 0,
							XmNmarginHeight, 0,
							XmNborderWidth, 0,
							NULL );
  frame = XtVaCreateWidget( "iconframe", xmFrameWidgetClass, board,
							XmNshadowThickness, 2,
							XmNborderWidth, 0,
							NULL );
  if( to_popout == False  ) {
	icon_popin();
  } else {
	icon_popout();
  }
  
  
  s = XmStringCreateSimple( myname  );
  
  label = XtVaCreateWidget( "iconlabel", xmLabelWidgetClass, frame,
							XmNlabelString, s ,
							XmNmarginLeft, 5,
							XmNmarginRight, 5,
							XmNmarginTop, 2,
							XmNmarginBottom, 2,
							XmNborderWidth, 0,
							NULL );
  
  XpmAttributes xpm_attr;
  xpm_attr.valuemask = XpmCloseness  | XpmExactColors;
  xpm_attr.exactColors = 0;
  xpm_attr.closeness = CLOSENESS;
  int res = XpmReadPixmapFile( XtDisplay( label), defaultpixmap ,
							   xpmfilename,
							   &pixmap, 0,
							   &xpm_attr );
  
  
  // debug
  switch( res ) {
  case XpmColorError:
	cout << "Toybox: color error during pixmap alloc\n" ;
	break;
  case XpmOpenFailed:
	cout << "Toybox: open failed for xpm file" << xpmfilename << endl ;
	break;
  case XpmFileInvalid:
	cout << "Toybox: " << xpmfilename <<" xpm file invalid\n";
	break;
  case XpmColorFailed:
	cout << "Toybox: color alloc failed for xpm-file " << xpmfilename << endl;
	break;
  }

  using_default = 0;
  if( (res != XpmSuccess) && (res !=XpmColorError) ) {
	pixmap = WM_Icon::defaultpixmap;
	using_default = 1;
  }
  
  pb = XtVaCreateWidget( "iconpb", xmPushButtonWidgetClass, board,
						 XmNlabelType, XmPIXMAP,
						 XmNlabelPixmap, pixmap,
						 XmNborderWidth, 0,
						 XmNuserData, (XtPointer) this,
						 NULL );


	  
  // could use the xpm Attributes but can't rely on it being set
  // - bug in libXpm ?
  Window root_ret;
  unsigned int bw_ret, depth_ret;
  int x_ret, y_ret;
  unsigned int wid, hei;
  if( XGetGeometry( display, pixmap, &root_ret, &x_ret, &y_ret,
					&wid, &hei, &bw_ret, &depth_ret ))
	XtVaSetValues( pb, XmNwidth, wid, XmNheight, hei , NULL );

  
  XpmFreeAttributes( &xpm_attr );
  XmStringFree( s );
  // free new'ed string space from icon_init
  delete [] myname;
  delete [] xpmfilename;
  
  // don't have ARM nor any good book on C++..donations, anyone??
  myname = 0;
  xpmfilename = 0;
  done = 1;
  
}


int  WM_Icon::is_unmapped()
{
  XWindowAttributes attr;
  
  if( XGetWindowAttributes( display, w, &attr )) {
	if(  attr.map_state == IsUnmapped ) return 1;
  }
  
  return 0;
}

void WM_Icon::icon_set_name( char *n )
{
  XmString s;
  
  s = XmStringCreateSimple( n );
  XtUnmanageChild(frame);
  XtVaSetValues( label, XmNlabelString, s, NULL );
  XmStringFree( s );
  icon_manage( 0, 0);
}

	 
void
WM_Icon::icon_add_activate_callback(XtCallbackProc proc,XtPointer client )
{
  XtAddCallback( pb, XmNactivateCallback, proc, client );
}
  
void WM_Icon::icon_manage(int x, int y )
{
  Dimension hei, wid;
  Dimension pb_hei, pb_wid;
  
  if( done ) {

	
	XtManageChild( label );
	XtVaGetValues( label, XmNwidth, &wid, XmNheight, &hei, NULL );
	XtVaSetValues( frame, XmNwidth, wid, XmNheight, hei , NULL );

	// frame is fully sized, not positioned
	
	XtManageChild(pb);	  
	XtVaGetValues( pb,
				   XmNheight, &pb_hei,
				   XmNwidth, &pb_wid,
				   NULL );
	XtUnmanageChild( pb );
	XtVaSetValues( pb, XmNy, 0 , NULL );
	
	// y-offset of frame
	XtVaSetValues( frame, XmNy, pb_hei+2, NULL );
	
	// determine x-offset of frame
	if( pb_wid < wid ) {
	  XtVaSetValues( frame, XmNx, 0, NULL );
	  XtVaSetValues( pb, XmNx, (wid-pb_wid)/2, NULL );
	} else {
	  XtVaSetValues( frame, XmNx, (pb_wid - wid)/2, NULL );
	  XtVaSetValues( pb, XmNx, 0, NULL );
	}
	XtManageChild( frame );
	XtManageChild( pb );
	icon_width = (wid>pb_wid) ? wid : pb_wid ;
	icon_height = pb_hei + hei+ 1;
	
	XtVaSetValues( board, XmNx, x, XmNy, y,
				   XmNwidth, icon_width,
				   XmNheight, icon_height,
				   NULL );
	XtManageChild( board );


  }
  
}

void WM_Icon::icon_bounding_box( int *x_ret,int *y_ret,int *wid,int *hei)
{
  Position x, y;
  XtVaGetValues( board, XmNx, &x, XmNy, &y , NULL );
  *wid = icon_width;
  *hei = icon_height;
  *x_ret = (int) x;
  *y_ret = (int) y;
}
void WM_Icon::icon_pb_bounding_box( int *x_ret,int *y_ret,int *wid,int *hei)
{
  Position x,y;
  Dimension w, h;
  XtVaGetValues( pb, XmNx, &x, XmNy, &y, XmNwidth, &w, XmNheight, &h, NULL);
  *wid = (int) w;
  *hei = (int) h;
  *x_ret = (int) x;
  *y_ret = (int) y;
}

void WM_Icon::icon_frame_bounding_box( int *x_ret,int *y_ret,int *wid,int *hei)
{
  Position x,y;
  Dimension w, h;
  XtVaGetValues( frame, XmNx, &x, XmNy, &y, XmNwidth, &w, XmNheight, &h, NULL);
  *wid = (int) w;
  *hei = (int) h;
  *x_ret = (int) x;
  *y_ret = (int) y;
}

	
void WM_Icon::icon_moveto( int x, int y )
{
  Position oldx, oldy;
  XtVaGetValues( board, XmNx, &oldx, XmNy, &oldy, NULL );
  if( (oldx != x ) || ( oldy != y )) {
	XtVaSetValues( board, XmNx, x, XmNy, y, NULL );
  }
}

void WM_Icon::icon_iconify()
{
  (void) XIconifyWindow( display, w, 0 );
  XFlush( display );
	 
}

void WM_Icon::icon_deiconify()
{
  XMapWindow( display, w );
  XFlush( display );
}

void WM_Icon::icon_popout()
{
  XtVaSetValues( frame,
				 XmNshadowType, XmSHADOW_OUT,
				 XmNshadowThickness, 2,
				 NULL );
}

void WM_Icon::icon_popin()
{
  XtVaSetValues( frame,
				 XmNshadowType, XmSHADOW_IN,
				 XmNshadowThickness, 2,
				 NULL );
}

WM_Icon::~WM_Icon()
{
  Pixmap pix;
  if( done  && !using_default) {
	XtVaGetValues( pb, XmNlabelPixmap, &pix, NULL );
	if( pix != XmUNSPECIFIED_PIXMAP ) {
	  XFreePixmap(display, pix );
	}
  }
  if( done ) XtDestroyWidget( board );
  if( myname  ) { delete [] myname; }
  if( xpmfilename ) { delete [] xpmfilename; }
  done =0;
}


// key: W, items: T*
template<class T ,class W>
class Pset {
  int numitems;
  T **items;
  W *keys;
public:
  Pset() {
	numitems = 0;
	items = 0;
	keys = 0;
  }
  ~Pset() {
	delete [] items;
	delete [] keys;
  }
  
  void insert( T *t , W w ) {
	// insert into Pset pointer to T
	int i;
	T **temp;
	W *ktemp;

	numitems++;
	
	temp = new T * [numitems];
	ktemp = new W  [numitems];
	
	for(i=0; i<numitems-1; i++) {
	  temp[i] = items[i];
	  ktemp[i] = keys[i];
	}
	
	temp[i] = t;
	ktemp[i] = w;

	delete [] items ;
	delete [] keys;
	items = temp;
	keys = ktemp;
  }

  T * operator[]( W w ) {
	if( exists( w ) ) {
	  return getvalue( w );
	} else {
	  cout << "Pset: no such index\n";
	  exit(0);
	  return 0;
	}
  }
  
  int exists( W w ) {
	int i;
	for(i=0; i<numitems; i++) {
	  if( keys[i] == w ) {
		return 1;
	  }
	}
	return 0;
  }
  
  T *  remove( W w ) {
	// remove from Pset pointer to T and return corresp. T *
	
	int i;
	int found=0, index=0;
	T *tp=0;
	
	for( i=0; i<numitems; i++) {
	  if( keys[i] == w ) {
		found = 1;
		index = i;
		tp = items[i];
	  }
	}
	
	if( found ) {
	  T **temp;
	  W *ktemp;

	  temp = new T * [--numitems];
	  ktemp = new W [numitems];

	  int c=0;
	  
	  for(i=0; i<numitems+1; i++ ) {
		
		if( i != index ) {
		  temp[c] = items[i];
		  ktemp[c] = keys[i];
		  c++;
		}
	  }
	  delete [] items;
	  delete [] keys;
	  items = temp;
	  keys = ktemp;
	  return tp;
	} else {
	  return 0;
	}
	
  }

  T *  getvalue( W w ) {
	int i;
	for( i=0; i<numitems; i++ ) {
	  if( keys[i]==w ) {
		return items[i];
	  }
	}
	cout << "Pset: no such key ";
	exit(0);
	return 0;
  }
  
  int getnumitems()  { return numitems; }
  T ** getcontents() { return items; }
  
};


/*  FUNCTION PROTOTYPES */

void add_input_handler( Window w, long event_mask, int add_icon );	
int no_transient( Window w );

void RemoveIcon( Widget w, XEvent e, String params, Cardinal num_params );
void ResizeAction( Widget w, XEvent e, String params, Cardinal num_params );

void RaiseWindow( Widget w, XEvent e, String params, Cardinal num_params );
void DoResize();
void IconIconifyCB( Widget w, XtPointer client, XtPointer call );

int has_name( Display *d, Window w, char *name );
int initially_iconic( Window w );
void external_event( XEvent event, Boolean &b );
void check_events( XtPointer client, XtIntervalId *id );
int still_around( Window w);
int still_around_deep( Window par, Window win );
void unmap_existing_icons( Window win );
int dont_show( char *wname );

void hide_window( Display *d, Window w );
void init_xtresource( char *rname, char *default_add);

void switch_x_y( XRectangle *xr, int n );
void switch_width_height( XRectangle *xr, int n );

void inv_x( XRectangle *xr, int n, unsigned short wid );
void inv_y( XRectangle *xr, int n, unsigned short hei );

void layout( XRectangle *xr, int n, unsigned short scr_wid,
			 unsigned short scr_hei );

void check_a_or_b( char *index, char *a, char *b, Direction aval,
				   Direction bval, Direction defaultval,Direction *direction );

// generic wrapper to XFree
void GFree( XWMHints *h );
void GFree( Window *h );
void GFree( unsigned char *h );
void GFree( char * h );

/* GLOBALS -COULD BE CLEANED UP */
static XtActionsRec actions[]={
  {"ResizeAction", (XtActionProc) ResizeAction },
  {"RaiseWindow", (XtActionProc) RaiseWindow },
  {"RemoveIcon", (XtActionProc) RemoveIcon },
  NULL
};

// must set this here or else g++ won't link
Pixmap WM_Icon::defaultpixmap = XmUNSPECIFIED_PIXMAP;

XtAppContext app_con;

Widget toplevel;
XEvent event ;
// disp used for our widget tree, sdisp used for intercepting external events
// had to be this way .. XtAppMainLoop() crashes when the associated
// display is used for other purposes...
Display *disp, *sdisp;
Window win, root;

#ifdef HAVE_SHAPE_EXTENSION
int float_on_root;
#endif

Window toybox_icon_window;
//set in unmap_existing_icons()

Widget bb, sw;
// additional translations for btn3
XtTranslations xttrans;

XtResource xtresource;

Atom atom;
// atom for WM_STATE

char default_xpmfile[200]= DEFAULT_XPM_FILE ;
char xpmfilename[200];
char *defaultPixmapFile;

/* startup dialog */
int n=0;
Arg args[10];
Widget dialog;
XmString ss, sd;

// Toybox logo - used throughout
Pixmap pix;

long std_mask;
char window_name[MAX_NAME_LENGTH+100];

char defaultlayoutpolicy[] = "top_left";
Direction direction;

Cursor watch_cursor;

int i;

char const * fallback_resources[] = {
  "Toybox*borderWidth: 0",
  "Toybox*bb*XmPushButton.width: 50",
  "Toybox*bb*XmPushButton.height: 50",
  NULL
};

Pset<WM_Icon, Window> iconbag;
Pset<Window, Window> iconwindowbag; // second arg=icon window from mwm


int main(int argc, char **argv)
{


  toplevel = XtAppInitialize(&app_con, "Toybox",
							 NULL, 0,
							 // on Suns this Cardinal *
							 // might be defined as int *
							 // - ignore it.
							 &argc, (char **)argv,
							 (char **) fallback_resources,
							 NULL, 0);
						   

  sw = XtVaCreateWidget("sw", xmScrolledWindowWidgetClass, toplevel,
						XmNscrollingPolicy, XmAUTOMATIC,
						XmNscrollBarDisplayPolicy, XmAS_NEEDED,
						XmNvisualPolicy, XmVARIABLE,
						XmNshadowThickness, 0,
						XmNspacing, 0,
						XmNborderWidth,0,
						XmNscrolledWindowMarginHeight, 0,
						XmNscrolledWindowMarginWidth, 0,
						NULL );
  
  bb = XtVaCreateWidget( "bb", xmRowColumnWidgetClass, sw,
						 XmNpacking, XmPACK_NONE,
						 XmNborderWidth, 0,
						 NULL );
  XtManageChild( sw );
  
  XtAppAddActions( app_con, actions, 3);
  XtAugmentTranslations( toplevel,
						 XtParseTranslationTable
						 (" <ConfigureNotify>: ResizeAction()\n"));

  disp = XtDisplay( toplevel );

  sdisp = XOpenDisplay( DisplayString( disp ) );
  // KNF
  if( sdisp == 0 ) {
	cout << "Toybox: internal error - couldn't make 2nd connection\n";
	exit(0);
  }
  
  win = XtWindow( toplevel );
  root = DefaultRootWindow( disp );
  xttrans = XtParseTranslationTable("!<Btn3Down> : RaiseWindow()\n\
Shift <Btn3Down> : RemoveIcon()\n");

  atom = (Atom) None;
  
  do {
	atom = XInternAtom( sdisp, "WM_STATE", True );
  } while ( atom == (Atom) None  ); // busy wait for window-manager
  
  // from the include file above
  if( XpmCreatePixmapFromData( disp, root, defaultpixmapdata,
							   &pix,0,0)!=XpmSuccess){
	cout << "Toybox: Radical! Couldn't even allocate 3 colours!\n";
	cout << " - bummer. Using stipple as default pixmap instead.\n";
	pix  = XmGetPixmap( XtScreen(toplevel), "25_foreground", 1,0);
  }

  WM_Icon::defaultpixmap = pix;
  // get user specified default-pixmapfile
  
  init_xtresource( "defaultPixmapFile", default_xpmfile );
  
  XtGetApplicationResources( toplevel, (XtPointer) &defaultPixmapFile,
							 &xtresource, 1, NULL, 0);
#ifdef HAVE_SHAPE_EXTENSION
  // find out if we need to float on root
  char *temp=0, *midp;

  init_xtresource( "floatOnRoot", "YES" );
  
  XtGetApplicationResources( toplevel, (XtPointer) &temp,
							 &xtresource, 1, NULL, 0);
  if( strcmp( temp, "NO" ) ) {
	float_on_root = 1;
  } else {
	float_on_root = 0;
  }

  int event_base, error_base;
  if( float_on_root &&
	  !XShapeQueryExtension( sdisp, &event_base, &error_base )) {
	cout << " The X-server " <<  DisplayString(sdisp ) \
         << " doesn't support the SHAPE extension.\n";
	cout << " Ignoring toybox*floatOnRoot: YES. Sorry!\n";
	float_on_root = 0;
  }
  // float on root doesn't change from here on.
  if( float_on_root ) {
	// don't need the scrolledwindow
	
	XtDestroyWidget( sw );
	
	bb = XtVaCreateWidget( "bb", xmRowColumnWidgetClass, toplevel,
						   XmNpacking, XmPACK_NONE,
						   XmNborderWidth, 0,
						   XmNwidth, WidthOfScreen(XtScreen(toplevel)),
						   XmNheight, HeightOfScreen(XtScreen(toplevel)),
						   NULL );

	// make mwm ignore our transaparent (soon to be) toplevel
	// don't make it appear just yet..map it later.
	XtVaSetValues( toplevel, XmNwidth,
				   WidthOfScreen(XtScreen(toplevel)),
				   XmNheight,
				   HeightOfScreen(XtScreen(toplevel)),
				   XmNx,0, XmNy, 0,
				   //XmNoverrideRedirect, True,
				   //XmNallowShellResize, True,
				   XmNmappedWhenManaged, False,
				   NULL);

  }
  // get layout policy
  init_xtresource( "layoutPolicy", defaultlayoutpolicy );
  temp = 0;
  XtGetApplicationResources( toplevel, (XtPointer) &temp,
							 &xtresource, 1, NULL, 0);

  // Thanks to Kenneth "A Law Office!" Flaxman for this tip on SYSV/BSD diff.
#ifdef SYSV
  midp = strchr( temp , '_' );
#else //BSD? 
  midp = index(temp, '_' );
#endif  
  
  if( midp ) {
	*midp = '\0';
	midp ++; // midp now points to char after '_'
	if( !strcmp( temp, "top" ) ) {

	  check_a_or_b( midp, "left", "right", top_left, top_right,
					top_left, &direction );
	  
	} else if( !strcmp( temp, "bottom")) {
	/* if midp=="left" then direction=2 else
       if midp=="right" then direction = 3 else direction = 0 */
	
	check_a_or_b( midp, "left", "right", bottom_left, bottom_right,
				  top_left, &direction );
	 
	} else if( !strcmp( temp, "left" )) {
	  check_a_or_b( midp, "top", "bottom", left_top, left_bottom,
					top_left, &direction);

	} else if( !strcmp( temp, "right")) {
	  check_a_or_b( midp, "top", "bottom", right_top, right_bottom,
					top_left, &direction );
	}
  } else {
	direction = top_left;
  }

#endif
  XtManageChild(bb );  
  
  // detect Resize and Property Changes 
  std_mask = SubstructureNotifyMask | PropertyChangeMask | ExposureMask;

  // need to grab server..don't want windows to unmap/destroy while
  // we're querying the tree
  
  XGrabServer( sdisp );
  
  watch_cursor = XCreateFontCursor( sdisp, XC_watch );
  n= XGrabPointer( sdisp, root, False, Button1MotionMask | ButtonPressMask |
				   PointerMotionMask | ButtonReleaseMask, GrabModeAsync,
				   GrabModeAsync, root, watch_cursor, CurrentTime );
  if( n == GrabSuccess ) {
	XFlush( sdisp );
  }
  
  unmap_existing_icons(root );
  
  add_input_handler( root, std_mask, True);


  WM_Icon **wp;
  wp = iconbag.getcontents();
  
  for( i=0; i < iconbag.getnumitems(); i++ ) {
	if( initially_iconic( wp[i]->get_window() ) ) {
	  wp[i]->set_popout_state( False );
	} else {
	  if( ! wp[i]->is_unmapped() ) {
		wp[i]->set_popout_state(True);
	  } else {
		wp[i]->set_popout_state( False );
	  }
	}
  }
  
  XUngrabServer( sdisp );
  XFlush( sdisp );
  
  if( n== GrabSuccess ) XUngrabPointer( sdisp, CurrentTime );

  XtRealizeWidget( toplevel );
  
#ifdef HAVE_SHAPE_EXTENSION
  XRectangle xre;
  xre.x = xre.y = 0;
  xre.width = xre.height = 0;
  if( float_on_root ) {
	// make whole toplevel transparent
	XShapeCombineRectangles( disp, XtWindow( toplevel ), ShapeBounding,
							0, 0, &xre, 1, ShapeSet, UNSORTED );
  }
#endif
  XtVaSetValues( toplevel, XmNmappedWhenManaged, True, NULL );
  XtMapWidget( toplevel );
  
  for( i=0; i < iconbag.getnumitems(); i++ ) {
	Window tw;
	tw = wp[i]->get_window();
	wp[i]->icon_realize();
	wp[i]->icon_augment_translations(xttrans);
	wp[i]->icon_add_activate_callback( (XtCallbackProc) IconIconifyCB,
									   (XtPointer)tw);
	wp[i]->icon_manage(0,0);
  }
  
  DoResize();
  
  ss = XmStringCreateLtoR( "Toybox release 1.3\n\
Written by: Tse Huong Choo\n\
E-mail suggestions to: \n\
thc2@doc.ic.ac.uk",
						   XmSTRING_DEFAULT_CHARSET );
  
  sd = XmStringCreateSimple("    Done    ");
  XtSetArg( args[n], XmNmessageString, ss ); n++;
  XtSetArg( args[n], XmNsymbolPixmap, pix); n++;
  XtSetArg( args[n], XmNmessageAlignment, XmALIGNMENT_CENTER );n++;
  XtSetArg( args[n], XmNokLabelString, sd ); n++;
  dialog = XmCreateInformationDialog( toplevel, "", args, n );
  XtUnmanageChild( XmMessageBoxGetChild( dialog, XmDIALOG_CANCEL_BUTTON));
  XtUnmanageChild( XmMessageBoxGetChild( dialog, XmDIALOG_HELP_BUTTON ));

  XtManageChild( dialog );
  XmStringFree( ss );
  XmStringFree( sd );
  

  XtIntervalId xid;
  xid = XtAppAddTimeOut( app_con, 50, (XtTimerCallbackProc) check_events,
						   0 );

  XtAppMainLoop( app_con);
  return 0;
}

void check_a_or_b( char *midp, char *a, char *b,Direction aval,Direction bval,
				   Direction defaultval, Direction *direction)
{
  if( !strcmp( midp, a )) {
	*direction = aval;
  } else if( !strcmp( midp, b)) {
	*direction = bval;
  } else {
	*direction = defaultval;
  }
}

void hide_window( Display *d, Window w )
// currently only unmapping the window...could try to move it
// out of sight instead.
{
  XUnmapWindow( d, w );
}

  
void unmap_existing_icons( Window win )
// if toybox crashes on startup on XUnmapWindow(), then
//	 change the way prop is indexed/dereferenced to give w.
{
  int i;
  Window root_ret, par_ret, *childs=0;
  unsigned nchild;
  Atom type_ret;
  int format_ret;
  unsigned long nitems, bytes;
  unsigned char *prop;
  Window w, *windp;
  int dosubtrees = 1;

  // should check if it's our toplevel window!

  
  if( XGetWindowProperty( sdisp, win, atom,
						  0, 10, False, AnyPropertyType,
						  &type_ret, &format_ret,
						  &nitems, &bytes, &prop )==Success) {

	if( nitems > 0) {
	  
	  // dodgy code - could someone with the ICCCM find
	  // out what the exact values should be?
	  // CARD32 equiv. unsigned long (on my Sparc IPX anyway...)
	  // WM_STATE property is suggested to be:
	  // Field     Type
	  // state     CARD32
	  // icon      WINDOW
	  // but contents of WM_STATE is not fixed yet .. will be
	  // specified fully in the coming Window and Session Manager
	  // Conventions Manual
	  // could do sizeof(CARD32) but would have to #include Xmd.h!
	  // so assume 'typedef unsigned long CARD32' as in Xmd.h
	  
	  w = *((Window *)(prop+ sizeof( unsigned long) ));

	  dosubtrees = 0;
	  
      // still_around() - added Wed May 26 09:57:42 BST 1993
	  
	  if( (w != (Window) None) && still_around_deep(root, w) ) {
 		int carryon=1;
		Window previous;
		// trace up the tree to root
		do {

		  if(XQueryTree( sdisp, w, &root_ret, &par_ret,&childs, &nchild)) {
			GFree( childs );
			previous = w;
			w = par_ret;
  		  } else carryon = 0; // XQueryTree() failed - do best we can
		} while ( carryon && (root_ret != par_ret) );

		if( root_ret == par_ret ) w = previous; // step down a level
		
		// if this is our window - needed for safe keeping
		if( win == XtWindow( toplevel )) {
		  toybox_icon_window = w;
		}
		
		// w's parent is now the root

		hide_window( sdisp, w );
		windp = new Window;
		*windp = w;
		iconwindowbag.insert( windp, w );
	  }
	}
	GFree( prop );
  }
  
  if(XQueryTree( sdisp, win, &root_ret, &par_ret,&childs, &nchild )
	 && dosubtrees ) {
	for( i=0; i < (int) nchild; i++ ) {
	  unmap_existing_icons( childs[i] );
	}
	GFree( childs );
  }
}




void check_events( XtPointer client, XtIntervalId *id )
{
  XEvent event;
  XtIntervalId xid;
  Boolean must_resize;

  
  must_resize = False;
  while( XCheckMaskEvent( sdisp, std_mask, &event )== (Bool)True) {

	external_event( event, must_resize );

  }

  if( must_resize == True ) DoResize();

  
  xid = XtAppAddTimeOut( app_con, TIME_OUT,
						 (XtTimerCallbackProc) check_events, 0 );
  
}

void external_event( XEvent event, Boolean &doresize )
{

  WM_Icon **wpp=0;
  Window dw,wi,wge,we,cw;
  int made_icons = 0;
  XPropertyEvent *xe;
  XWMHints *hi;
  WM_Icon *wii;
  
  switch( event.type ) {
	
  case DestroyNotify:
	
	dw = ((XDestroyWindowEvent *)&event)->window;
	if( iconbag.exists( dw ) ) {

	  delete iconbag[dw];
	  
	  (void ) iconbag.remove( dw );
	  doresize=True;
	  
	} else if( iconwindowbag.exists( dw )) {
	  
	  delete iconwindowbag[dw];
	  
	  (void) iconwindowbag.remove( dw );
	}
	
	break;
  case UnmapNotify:

	if( iconbag.exists(((XUnmapEvent *)&event)->window)) {
	  iconbag[ ((XUnmapEvent *)&event)->window ]->icon_popin();
	} 
	
	break;
	
  case GraphicsExpose:

	
	wge = (Window) ((XGraphicsExposeEvent*)&event)->drawable;
	
	if( iconbag.exists( wge )) {
	  iconbag[wge]->icon_popout();
	}
	break;
  case Expose:

	
	we = ((XExposeEvent *)&event)->window;
	
	if( iconbag.exists( we )) {
	  iconbag[we]->icon_popout();
	}
	break;
  case CreateNotify:

	
	cw = ((XCreateWindowEvent *)&event)->window;
	if( (((XCreateWindowEvent *)&event)->override_redirect != True)) {

	  // may be too late already
	  XGrabServer( sdisp );  
	  XFlush( sdisp );
	  
	  if( still_around( cw) ) {

		add_input_handler( cw, std_mask, True);

		// delete its icon, if needed. Also insert into iconwindowbag
		unmap_existing_icons( cw );
		// say if we need to popout

		wpp = iconbag.getcontents();
		for( i=0; i<iconbag.getnumitems(); i++ ) {
		  if( !(wpp[i]->is_done())) {
			made_icons = 1;
			if(initially_iconic(wpp[i]->get_window() ) ) {
			  wpp[i]->set_popout_state( False );
			} else {
			  if( ! wpp[i]->is_unmapped() ) {
				wpp[i]->set_popout_state(True);
			  } else {
				wpp[i]->set_popout_state( False );
			  }
			}
		  }
		}
	  }
	  XUngrabServer( sdisp );
	  XFlush( sdisp );
      // no problems with un-synched sdisp now
	  if( made_icons ) {

		for(i=0; i<iconbag.getnumitems() ; i++ ) {
		  if( !(wpp[i]->is_done())) {
			Window tw;
			tw = wpp[i]->get_window();
			wpp[i]->icon_realize();
			wpp[i]->icon_augment_translations(xttrans);
			wpp[i]->icon_add_activate_callback( (XtCallbackProc) IconIconifyCB,
												(XtPointer)tw);
			wpp[i]->icon_manage(0,0);
		  }
		}
		doresize = True;
	  }
	}
	break;
  case MapNotify:

	wi = ((XMapEvent *)&event)->window;

	if( iconbag.exists( wi ) ) {
	  iconbag[wi]->icon_popout();
	  
	} else if( iconbag.exists( ((XMapEvent *)&event)->event )) {
	  
	  iconbag[((XMapEvent *)&event)->event]->icon_popout();
	} else if( (wi!=toybox_icon_window) &&iconwindowbag.exists( wi )) {
	  // iconwindow was mapped
	  XGrabServer( sdisp );

	  hide_window( sdisp, wi );
	  XUngrabServer( sdisp );
	}
					
	break;
  case PropertyNotify:
	
	xe = (XPropertyEvent *) &event;
	hi = 0;


	if( xe->atom == XA_WM_NAME) {

	  // filemgr crashes here!?
	  XGrabServer( sdisp );
	  XFlush( sdisp );



	  if( !(iconbag.exists( xe->window)) &&
		  still_around( xe->window) &&
		  has_name( sdisp, xe->window, window_name) &&
		  !dont_show(window_name) ) {
		
		char *filename;

		init_xtresource( window_name, defaultPixmapFile );
		XtGetApplicationResources( toplevel, (XtPointer) &filename,
								   &xtresource, 1, NULL, 0);
		
		wii = new WM_Icon;
		iconbag.insert( wii, xe->window );
		wii->icon_init( sdisp, xe->window, window_name,bb,filename);
		if(initially_iconic(xe->window))wii->set_popout_state(False);
		if( !(wii->is_unmapped())) wii->set_popout_state( True );

		XUngrabServer( sdisp );
		XFlush( sdisp );
		
		Window tw;
		tw = wii->get_window();
		wii->icon_realize();
		wii->icon_augment_translations( xttrans );
		wii->icon_add_activate_callback((XtCallbackProc) IconIconifyCB,
										(XtPointer)tw);
		
		wii->icon_manage( 0, 0);
		
		doresize= True;

	  } else {
		// existing window changed its name - sdisp still grabbed
		// NOTE: IF new name is one defined as DONT_SHOW, then what??
		// currently, icon is removed .
		
		if( has_name(sdisp, xe->window, window_name ) &&
			iconbag.exists( xe->window ) ) {
		  if( !dont_show( window_name )) {
			XUngrabServer( sdisp );
			XFlush( sdisp );
			
			iconbag[xe->window]->icon_set_name( window_name );

			doresize = True;
		  } else if( iconbag.exists( xe->window )) {
			// name defined as dont_show and one of our windows
			
			delete iconbag[xe->window];
			(void) iconbag.remove( xe->window );
			doresize = True;
		  }
		  
		} else {
		  XUngrabServer( sdisp );
		  XFlush( sdisp );
		}
		
	  }

	} else if( (xe->atom == XA_WM_TRANSIENT_FOR) &&
			   (xe->state == PropertyNewValue)) {
	  // setting the TRANSIENT_FOR property , must be
	  // a dialog box, so destroy icon
	  
	  if( iconbag.exists( xe->window ) ) {
		
		delete iconbag[xe->window];
		(void ) iconbag.remove( xe->window );
		
		doresize = True;
	  }
			
		
	} else if( xe->atom == XA_WM_HINTS ) {
	  
	  hi = 0;

	  XGrabServer( sdisp );
	  XFlush( sdisp );
	  
	  if( still_around( xe->window ) ) {
		hi = XGetWMHints( sdisp, xe->window );
	  }
	  XUngrabServer( sdisp );
	  XFlush( sdisp );

	  if( hi &&(hi->flags&StateHint) ) {
		
		if( hi->initial_state == NormalState ) {
		  if( iconbag.exists( xe->window ) ) {
			
			iconbag[xe->window]->icon_popout();
		  }
		} else {
		  if( iconbag.exists( xe->window ) ) {
			
			iconbag[xe->window]->icon_popin();
		  }
		}
		
	  }

	  
	  if( hi && (hi->flags&WindowGroupHint) &&
		  (hi->window_group != (XID) xe->window)) {
		
		if( iconbag.exists( xe->window )) {
		  
		  delete iconbag[xe->window];
		  (void ) iconbag.remove( xe->window );
		  doresize = True;
		}
	  }
	  if( hi ) GFree( hi );
	} else if( xe->atom == atom ) {
	  // WM_STATE
	  if( xe->state == PropertyNewValue) {
		// icon window being set
		XGrabServer( sdisp );
		if( still_around( xe->window )) {
		  unmap_existing_icons( xe->window );
		} 
		XUngrabServer( sdisp );
		
	  } // else check for PropertyDelete and delete from
	  // iconwindowbag - this is a memory leak at the moment.
	  // It's not expected that iconwindows change during
	  // a window's lifetime. 
	}
	break;
  } // switch
}

void add_input_handler( Window w, long event_mask , int add_icon)
// assume XGrabServer( sdisp ) is in effect	 
{
  int i;
  Window root_ret, par_ret, *childs=0;
  unsigned nchild;
  char wname[MAX_NAME_LENGTH+20]; 
  char *filename; 
  int dosubtrees = 1; // do subtrees unless we need to make icon here
  XWMHints *hi=0;
  
  if( w == XtWindow( toplevel ) ) return;
  hi = XGetWMHints( sdisp, w );

  XSelectInput( sdisp, w, event_mask );
  
  if( add_icon ) {
	if( has_name( sdisp, w, wname ) && !dont_show(wname ) &&
		!( hi && (hi->flags & WindowGroupHint) &&
		   (hi->window_group != (XID) w ))) {

		
	  if( (!(iconbag.exists( w )))  &&  no_transient(w)) {
		// doesn't already exist in iconbag and hasn't
		// got the WM_TRANSIENT_FOR property

		init_xtresource( wname, defaultPixmapFile );
		XtGetApplicationResources( toplevel, (XtPointer) &filename,
								   &xtresource, 1, NULL, 0);
		
		WM_Icon *wp;
		
		wp = new WM_Icon;
		wp->icon_init( sdisp, w, wname, bb, filename );
		iconbag.insert( wp , w );

		dosubtrees = 0;
	  }
	}
  }
  
  if( hi )  GFree( hi );
  
  if ( dosubtrees &&
	   XQueryTree( sdisp, w, &root_ret, &par_ret,&childs, &nchild ) ) {

	for( i=0; i < (int) nchild; i++ ) {
	  add_input_handler( childs[ i ], event_mask, add_icon );
	}
	if( childs ) GFree( childs );
  }

}

int dont_show( char *wname )
{
  char *filename;
  
  init_xtresource( wname, defaultPixmapFile);
  XtGetApplicationResources( toplevel, (XtPointer) &filename,
							 &xtresource, 1, NULL, 0);
  

  if( !strcmp( DONT_SHOW, filename ) ) {
	return 1;
  }

  return 0;
}

void init_xtresource( char *rname , char *default_add)
{
  xtresource.resource_name = rname;
  xtresource.resource_class = XtCString;
  xtresource.resource_type = XtRString;
  xtresource.resource_size = sizeof( String );
  xtresource.resource_offset = 0;
  xtresource.default_type = XtRImmediate;
  xtresource.default_addr = default_add;
}


int initially_iconic( Window w )
{
  XWMHints *h=0;

  h = XGetWMHints( sdisp, w );
  if( h  ) {
	if( h->flags & StateHint ) {
	  if( h->initial_state != NormalState ) {
		// should've checked for IconicState instead
		// or should've called this initially_not_normal(). Duh!
		GFree( h );
		return 1;
	  }
	}
	GFree( h );
  }
  return 0;
}

// some systems (eg DEC) define XFree( XPointer )
// whereas Sparcs use XFree( void * )
// all GFree()s below may have to be #ifdef-ed.

void GFree( XWMHints *h )
{
  char *v;
  v = (char *) h;
  XFree( v );
}
void GFree( Window *h )
{
  char *v;
  v = (char *) h;
  XFree( v );
}

void GFree( unsigned char *h )
{
  char *v;
  v = (char *) h;
  XFree( v );
}

void GFree( char * h )
{
  XFree( h );
}


int no_transient( Window w )
{
  Window w_ret;
  if( XGetTransientForHint( sdisp, w, &w_ret )) {
	return 0;
  } else {
	return 1;
  }
}

  
void IconIconifyCB( Widget w, XtPointer client, XtPointer call )
{
  Window win = (Window) client; // note cast - more ANSI duh
  XGrabServer( sdisp );


  // must grab server - window could've been killed by now

  if( iconbag[win]->is_unmapped() ) {
	XUngrabServer( sdisp );
	iconbag[win]->icon_deiconify();
  } else {
	XUngrabServer( sdisp );
	iconbag[win]->icon_iconify();
  }

}

int still_around_deep( Window par, Window win )
{
  // deep version of still_around()
  Window root_ret, par_ret, *childs=0;
  unsigned nchild;
  int p;

  if( XQueryTree( sdisp, par,
				  &root_ret, &par_ret,&childs, &nchild ) ) {
	for( p=0; p < (int) nchild; p++ ) {
	  if( childs[p] == win ) {
		if( childs ) GFree( childs );
		return 1;
	  } 
	}
	// couldn't find it at this level
	for( p=0; p < (int) nchild; p++ ) {
	  if( still_around_deep( childs[p], win )) {
		// found it
		if( childs) GFree( childs );
		return 1;
	  }
	}
	// couldn't find it at all!

  }
  return 0;
}
  
int still_around( Window w)
// assume XGrabServer(sdisp) was called first!
{
  Window root_ret, par_ret, *childs=0;
  unsigned nchild;
  int p;
  
  if( XQueryTree( sdisp, root,
				  &root_ret, &par_ret,&childs, &nchild ) ) {
	
	for( p=0; p < (int) nchild; p++ ) {
	  if( childs[p] == w ) {
		if( childs ) GFree( childs );
		return 1;
	  } 
	}
	
	if( childs  ) GFree( childs );
  }
  return 0;
  
}

void RemoveIcon( Widget w, XEvent e, String params, Cardinal num_params )
{
  Window wind;
  XtPointer p;
  WM_Icon *i;
  int n;
  Window **window;
  XSetWindowAttributes attr;
  unsigned long mask;
  
  XtVaGetValues( w, XmNuserData, &p, NULL );
  i = (WM_Icon *) p;
  
  wind = i->get_window();
  if( iconbag.exists( wind )) {
	delete iconbag[wind];

	(void) iconbag.remove(wind);
	DoResize();
  }

  // exit if no more icons
  if( iconbag.getnumitems() == 0) {
	window =iconwindowbag.getcontents();
	attr.override_redirect = True;
	mask = CWOverrideRedirect;
	for(n =0; n < iconwindowbag.getnumitems(); n++) {
	  
	  // hack - XMapWindow() doesn't map unless its true according to man pages

	  attr.override_redirect = True;
	  XChangeWindowAttributes( sdisp, *(window[n]), mask, &attr);
	  
	  XMapWindow( sdisp, *(window[n]));
	  attr.override_redirect = False;
	  XChangeWindowAttributes( sdisp, *(window[n]), mask, &attr);
	}
	XSync( sdisp, False );
	exit(0);
  }
}
	
void RaiseWindow( Widget w, XEvent e, String params, Cardinal num_params )
{
  XtPointer p;
  Window wind;
  
  XtVaGetValues( w /* pb in an icon */, XmNuserData, &p, NULL );
  
  wind = ((WM_Icon *)p)->get_window();

  if( iconbag.exists( wind ) ) {
	XMapRaised( sdisp, wind );
	XFlush( sdisp );
  }
}

void DoResize()
{
  unsigned short scr_wid, scr_hei;
  XRectangle *xr, *il;
  int n;
  WM_Icon **wpp;
  int x,y,wid,hei;
  Dimension dwid, dhei;
  
  scr_wid=(unsigned short) WidthOfScreen( XtScreen(toplevel));
  scr_hei=(unsigned short) HeightOfScreen( XtScreen(toplevel));
  n = iconbag.getnumitems();
  
  wpp = iconbag.getcontents();
  
  xr = new XRectangle [ n ] ;
#ifdef HAVE_SHAPE_EXTENSION
  if( float_on_root ) {
	il = new XRectangle [ n*2 ];
  } else {
	XtVaGetValues( toplevel, XmNwidth, &dwid, XmNheight, &dhei, NULL );
	if( dwid > dhei ) {
	  //top left
	  direction = top_left;
	} else {
	  //right top
	  direction = left_top;
	}	
	scr_wid = (unsigned short ) dwid;
	scr_hei = (unsigned short) dhei;
  }
#else
  
  XtVaGetValues( toplevel, XmNwidth, &dwid, XmNheight, &dhei, NULL );
  if( dwid > dhei ) {
	//top left
	direction = top_left;
  } else {
	direction = left_top;
  }	
  scr_wid = (unsigned short ) dwid;
  scr_hei = (unsigned short) dhei;

#endif
  
  for(i=0; i<n; i++) {
	wpp[i]->icon_bounding_box(&x, &y, &wid, &hei);
	xr[i].width = (unsigned short) wid;
	xr[i].height = (unsigned short) hei;

#ifdef HAVE_SHAPE_EXTENSION
	if( float_on_root ) {
	  wpp[i]->icon_pb_bounding_box( &x, &y, &wid, &hei );
	  il[i*2].x = (short) x;
	  il[i*2].y = (short) y; 
	  il[i*2].width = (unsigned short) wid;
	  il[i*2].height = (unsigned short) hei;
	  
	  wpp[i]->icon_frame_bounding_box( &x, &y, &wid, &hei );
	  il[i*2+1].x = (short) x;
	  il[i*2+1].y = (short) y; 
	  il[i*2+1].width = (unsigned short) wid;
	  il[i*2+1].height = (unsigned short) hei;
	}
#endif	
  }

  if( direction==left_top || direction==left_bottom ||
	  direction==right_top || direction==right_bottom) {
	switch_width_height( xr, n );
	layout( xr, n, scr_hei, scr_wid );
  } else {
	layout( xr, n, scr_wid, scr_hei );
  }
  
  if( direction==left_top || direction==left_bottom ||
	  direction==right_top || direction==right_bottom ) {
	switch_width_height( xr, n );
	switch_x_y( xr, n );
  }
  
  switch( direction ) {
  case top_right: 
	inv_x(xr, n, scr_wid );
	break;
  case bottom_left: 
	inv_y(xr, n, scr_hei);
	break;
  case bottom_right: 
	inv_x(xr,n, scr_wid ); inv_y(xr,n, scr_hei);
	break;
  case left_bottom: 
	inv_y(xr, n, scr_hei);
	break;
  case right_top:
	inv_x(xr, n, scr_wid);
	break;
  case right_bottom:
	inv_x(xr,n, scr_wid); inv_y(xr, n, scr_hei );
	break;
  }

  XtUnmanageChild( bb );
  for( i=0; i<n; i++ ) {
	wpp[i]->icon_moveto( (int) xr[i].x, (int) xr[i].y );
#ifdef HAVE_SHAPE_EXTENSION	
	// shift label/pb rects by x,y of bboard
	if( float_on_root ) {
	  il[i*2].x += xr[i].x;
	  il[i*2].y += xr[i].y;
	  il[i*2+1].x += xr[i].x;
	  il[i*2+1].y += xr[i].y;
	}
#endif	
  }
  XtVaSetValues( bb, XmNwidth, WidthOfScreen(XtScreen(toplevel)),
				 XmNheight, HeightOfScreen( XtScreen(toplevel)),
				 NULL );
  XtManageChild( bb );

#ifdef HAVE_SHAPE_EXTENSION
  if( float_on_root ) {

	XShapeCombineRectangles( disp, XtWindow(toplevel), ShapeBounding,
							   0, 0, il, n*2,
							   ShapeSet, UNSORTED );
	
	int dest_y,dest_x;
	Window child_ret;
	  
	if(XTranslateCoordinates( disp, XtWindow(toplevel),XtWindow(bb),
							  0, 0, &dest_x, &dest_y, &child_ret )) {
	  if( dest_x != 0 || dest_y != 0 ) {
		XShapeOffsetShape( disp, XtWindow(toplevel), ShapeBounding,
						   -dest_x, -dest_y );
	  }
	}
  }

#endif
  
  delete [] xr;
  
#ifdef HAVE_SHAPE_EXTENSION  
  if( float_on_root ) delete [] il;
#endif  
  
}

void switch_x_y( XRectangle *xr, int n )
{
  int i;
  short temp;
  for(i=0; i<n; i++) {
	temp = xr[i].x;
	xr[i].x = xr[i].y;
	xr[i].y = temp;
  }
}

void switch_width_height( XRectangle *xr, int n )
{
  int i;
  unsigned short temp;
  for(i=0; i<n; i++) {
	temp = xr[i].width;
	xr[i].width = xr[i].height;
	xr[i].height = temp;
  }
}

void inv_x( XRectangle *xr, int n, unsigned short wid )
{
  int i;
  short s = (short) wid;
  for(i=0; i<n; i++) {
	xr[i].x = s - xr[i].x - xr[i].width;

  }
}

void inv_y( XRectangle *xr, int n, unsigned short hei )
{
  int i;
  short s = (short) hei;
  for(i=0; i<n; i++) {
	xr[i].y = s - xr[i].y - xr[i].height;

  }
}

void layout( XRectangle *xr, int n, unsigned short scr_wid,
			 unsigned short scr_hei )
{
  int i;
  Position xpos=BB_SPACING, ypos=BB_SPACING;
  unsigned short maxhei=0;
  WM_Icon **wpp;

  
  wpp = iconbag.getcontents();
  
  for( i=0; i<n; i++ ) {
	if( xr[i].height > maxhei ) maxhei = xr[i].height;
	
	if( (xpos +  xr[i].width + BB_SPACING ) < scr_wid ) {
	  xpos +=  xr[i].width + BB_SPACING;
	} else {
	  // run out of horiz space, start new row
	  xpos = xr[i].width + 2*BB_SPACING;
	  ypos += maxhei + BB_SPACING;
	}
	xr[i].x = (short ) (xpos- xr[i].width-BB_SPACING );
	xr[i].y = (short) ypos;
  }

}

void ResizeAction( Widget w, XEvent e, String params, Cardinal num_params )
{
  DoResize();
}



int has_name( Display *d, Window w, char *name )
{
  XTextProperty tprop;
  char **slist;
  int n;
  char *temp;
  temp = 0;

  if( !XFetchName( d, w, &temp ) ) {

	if( XGetWMName( sdisp, w, &tprop ) ) {
	  if( (tprop.format==8) && (tprop.encoding==XA_STRING) &&
		  XTextPropertyToStringList( &tprop, &slist, &n ) ) {
		(void) strncpy( name, slist[0], MAX_NAME_LENGTH-1 );
		// free char * list 
		XFreeStringList( slist );
		return 1;
	  } else {
		return 0;
	  }

	} else {

	  return 0;
	}
  } else {
	if( strlen( temp ) != 0) {
	  (void) strncpy( name, temp, MAX_NAME_LENGTH-1);
	  if( strlen( temp )>= MAX_NAME_LENGTH-1) {
		// terminate our name
		name[ MAX_NAME_LENGTH-1 ] = '\0';
	  }
	  if( temp ) GFree( temp );

	  return 1;

	} else {
	  if( temp ) GFree( temp );

	  return 0;
	}
  }

}



